# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

from abc import abstractmethod
from typing import Optional


class DeviceFlasherBase:
    @staticmethod
    @abstractmethod
    def flash(device_port, device_name, image_path, progress_callback=None):
        pass

    @staticmethod
    def get_boot_description(device_name: str) -> Optional[str]:
        return None



# Copyright (c) Acconeer AB, 2023-2024
# All rights reserved

from __future__ import annotations

import enum
from typing import Any, Optional, Tuple

import attrs
import h5py
import numpy as np
import numpy.typing as npt

from acconeer.exptool import a121, opser
from acconeer.exptool.a121.algo import (
    APPROX_BASE_STEP_LENGTH_M,
    ENVELOPE_FWHM_M,
    AlgoConfigBase,
    Controller,
)
from acconeer.exptool.a121.algo.presence._processors import (
    Processor,
    ProcessorConfig,
    ProcessorContext,
    ProcessorResult,
)


class DetectionState(enum.Enum):
    """Indicating the current detection state."""

    NO_DETECTION = enum.auto()
    """No detection."""

    DETECTION = enum.auto()
    """Detection."""

    RETENTION = enum.auto()
    """No detection. Retaining detection."""


@attrs.mutable(kw_only=True)
class ExampleAppConfig(AlgoConfigBase):
    # Setup parameters
    sensor_to_water_distance: float = attrs.field(default=0.12)
    """Distance from the sensor location to the center of the water jet (m)."""

    water_jet_width: float = attrs.field(default=0.02)
    """Water jet width (m).

    Defines the width of a region, centered around the water jet, where the hand motion is not
    analyzed, in order to exclude the water jet motion from the motion metric.
    """

    measurement_range_end: float = attrs.field(default=0.2)
    """Distance between sensor and end point of measurement range (m)."""

    # Filtering parameters
    filter_time_const: float = attrs.field(default=0.3)
    """Time constant of filter applied during processing.

    A lower value yield a more responsive algorithm. A higher value yield a more robust response.
    """

    threshold: float = attrs.field(default=1.5)
    """Threshold applied to the calculate metric.

    If the metric is above the threshold, a hand movement has been detected.
    """

    detection_retention_duration: float = attrs.field(default=1.0)
    """Detection retention duration (s).

    The detection will be kept as True for this duration, after detection is lost.
    """

    # Sensor config parameters
    hwaas: int = attrs.field(default=32)
    """HWAAS."""

    sweeps_per_frame: int = attrs.field(default=64)
    """Sweeps per frame."""

    sweeps_rate: int = attrs.field(default=1000)
    """Sweep rate."""

    frame_rate: int = attrs.field(default=10)
    """Frame rate."""

    def _collect_validation_results(self) -> list[a121.ValidationResult]:
        validation_results: list[a121.ValidationResult] = []

        return validation_results


opser.register_json_presentable(ExampleAppConfig)


@attrs.frozen(kw_only=True)
class ExtraResult:
    history: npt.NDArray[np.float64] = attrs.field()
    """Detection history."""

    history_time: npt.NDArray[np.float64] = attrs.field()
    """Time vector for detection history."""

    threshold: float = attrs.field()
    """Threshold."""


@attrs.frozen(kw_only=True)
class ExampleAppResult:
    detection_state: DetectionState = attrs.field()
    """The state of the detection."""

    extra_result: ExtraResult = attrs.field()
    """Extra result used for plotting."""


class ExampleApp(Controller[ExampleAppConfig, ExampleAppResult]):
    """Hand motion detection example app.

    :param client: Client
    :param sensor_id: Sensor id
    :param example_app_config: Hand motion detection configuration
    """

    processor: Processor
    HISTORY_LENGTH_S = 10.0

    def __init__(
        self,
        *,
        client: a121.Client,
        sensor_id: int,
        example_app_config: ExampleAppConfig,
    ) -> None:
        super().__init__(client=client, config=example_app_config)
        self.sensor_id = sensor_id
        self.config = example_app_config

        self.detection_retention_duration = int(
            round(example_app_config.detection_retention_duration * example_app_config.frame_rate)
        )

        self.started = False

    def start(
        self, recorder: Optional[a121.Recorder] = None, _algo_group: Optional[h5py.Group] = None
    ) -> None:
        if self.started:
            msg = "Already started"
            raise RuntimeError(msg)

        self._reinitialize_state_variables()

        sensor_config = self._get_sensor_config(self.config)
        self.session_config = a121.SessionConfig(
            {self.sensor_id: sensor_config},
            extended=False,
        )

        metadata = self.client.setup_session(self.session_config)
        assert isinstance(metadata, a121.Metadata)

        self.processor = Processor(
            sensor_config=sensor_config,
            processor_config=self._get_processor_config(self.config),
            context=ProcessorContext(estimated_frame_rate=self.config.frame_rate),
            metadata=metadata,
        )

        if _algo_group is None and isinstance(recorder, a121.H5Recorder):
            _algo_group = recorder.require_algo_group("faucet")

        if _algo_group is not None:
            _record_algo_data(_algo_group, self.sensor_id, self.config)

        if recorder is not None:
            self.client.attach_recorder(recorder)

        self.client.start_session()

        self.started = True

    def _reinitialize_state_variables(self) -> None:
        num_points_history = int(self.HISTORY_LENGTH_S * self.config.frame_rate)
        self.history = np.zeros(num_points_history)
        self.history_time = np.linspace(-self.HISTORY_LENGTH_S, 0, num_points_history)
        self.has_detected = False
        self.update_index = 0

    def get_next(self) -> ExampleAppResult:
        if not self.started:
            msg = "Not started"
            raise RuntimeError(msg)

        result = self.client.get_next()
        assert isinstance(result, a121.Result)

        assert self.processor is not None
        processor_result = self.processor.process(result)

        return self._process_processor_result(processor_result, self.config)

    def _process_processor_result(
        self, result: ProcessorResult, config: ExampleAppConfig
    ) -> ExampleAppResult:
        max_presence_score = max(result.inter_presence_score, result.intra_presence_score)

        # Determine detections state.
        detection_state = DetectionState.NO_DETECTION

        if result.presence_detected:
            self.has_detected = True
            self.update_index_at_detection = self.update_index
            detection_state = DetectionState.DETECTION

        if (
            self.has_detected
            and detection_state is not DetectionState.DETECTION
            and (self.update_index - self.update_index_at_detection)
            < self.detection_retention_duration
        ):
            detection_state = DetectionState.RETENTION

        # Prepare extra result(used for plotting).
        self.history = np.roll(self.history, shift=-1, axis=0)
        self.history[-1] = max_presence_score
        extra_result = ExtraResult(
            history=self.history,
            history_time=self.history_time,
            threshold=config.threshold,
        )

        self.update_index += 1

        return ExampleAppResult(detection_state=detection_state, extra_result=extra_result)

    def stop_detector(self) -> Any:
        if not self.started:
            msg = "Already stopped"
            raise RuntimeError(msg)

        self.client.stop_session()
        self.started = False

        return None

    def stop_recorder(self) -> Any:
        recorder = self.client.detach_recorder()
        if recorder is None:
            recorder_result = None
        else:
            recorder_result = recorder.close()

        return recorder_result

    def stop(self) -> Any:
        self.stop_detector()
        recorder_result = self.stop_recorder()

        return recorder_result

    @classmethod
    def _get_sensor_config(cls, config: ExampleAppConfig) -> a121.SensorConfig:
        """Convert example app config parameters to sensor config.

        Creates a subsweep configuration with
        - A subsweep between the sensor and the water jet(if it fits).
        - A subsweep between the water jet and the end of the measurement range(if it fits).
        """
        profile = a121.Profile.PROFILE_1
        step_length = 6
        water_jet_half_width = config.water_jet_width / 2.0

        subsweep_configs = []

        # Determine if subsweep between sensor and water jet is feasible
        if ENVELOPE_FWHM_M[profile] < (config.sensor_to_water_distance - water_jet_half_width):
            subsweep_config = a121.SubsweepConfig()
            subsweep_config.start_point = int(
                round(
                    (
                        config.sensor_to_water_distance
                        - ENVELOPE_FWHM_M[profile]
                        - water_jet_half_width
                    )
                    / APPROX_BASE_STEP_LENGTH_M
                )
            )
            subsweep_config.num_points = 1
            subsweep_config.step_length = step_length
            subsweep_config.profile = profile
            subsweep_config.hwaas = config.hwaas
            subsweep_config.receiver_gain = 4
            subsweep_configs.append(subsweep_config)

        # Determine if subsweep after water jet is required
        if (
            ENVELOPE_FWHM_M[profile] + water_jet_half_width + config.sensor_to_water_distance
            < config.measurement_range_end
        ):
            start_m = (
                config.sensor_to_water_distance + water_jet_half_width + ENVELOPE_FWHM_M[profile]
            )
            start_point = int(round(start_m / APPROX_BASE_STEP_LENGTH_M))
            num_points = max(
                1,
                int(
                    round(
                        (config.measurement_range_end - start_m)
                        / (step_length * APPROX_BASE_STEP_LENGTH_M)
                    )
                ),
            )

            subsweep_config = a121.SubsweepConfig()
            subsweep_config.start_point = start_point
            subsweep_config.num_points = num_points
            subsweep_config.step_length = step_length
            subsweep_config.profile = profile
            subsweep_config.hwaas = config.hwaas
            subsweep_config.receiver_gain = 4
            subsweep_configs.append(subsweep_config)

        sensor_config = a121.SensorConfig(subsweeps=subsweep_configs)
        sensor_config.frame_rate = config.frame_rate
        sensor_config.sweep_rate = config.sweeps_rate
        sensor_config.sweeps_per_frame = config.sweeps_per_frame
        sensor_config.inter_frame_idle_state = a121.IdleState.SLEEP
        sensor_config._inter_sweep_idle_state = a121.IdleState.SLEEP

        return sensor_config

    @classmethod
    def _get_processor_config(cls, config: ExampleAppConfig) -> ProcessorConfig:
        return ProcessorConfig(
            intra_enable=True,
            inter_enable=True,
            intra_detection_threshold=config.threshold,
            inter_detection_threshold=config.threshold,
            inter_phase_boost=False,
            phase_adaptivity_tc=5,
            inter_frame_presence_timeout=None,
            inter_frame_fast_cutoff=3.0,
            inter_frame_slow_cutoff=0.8,
            inter_frame_deviation_time_const=0.5,
            inter_output_time_const=config.filter_time_const,
            intra_frame_time_const=0.15,
            intra_output_time_const=config.filter_time_const,
        )


def _record_algo_data(
    algo_group: h5py.Group,
    sensor_id: int,
    config: ExampleAppConfig,
) -> None:
    algo_group.create_dataset(
        "sensor_id",
        data=sensor_id,
        track_times=False,
    )
    app_config_group = algo_group.create_group("example_app_config")
    opser.serialize(config, app_config_group)


def _load_algo_data(algo_group: h5py.Group) -> Tuple[int, ExampleAppConfig]:
    sensor_id = int(algo_group["sensor_id"][()])
    config = opser.deserialize(algo_group["example_app_config"], ExampleAppConfig)
    return sensor_id, config



# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

from __future__ import annotations

import logging
import shutil
import traceback
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Optional

from requests import Response, Session
from requests.cookies import RequestsCookieJar

from PySide6.QtCore import QThread, Signal

from acconeer.exptool._core.communication.comm_devices import CommDevice
from acconeer.exptool.app.new._exceptions import HandledException
from acconeer.exptool.flash import (
    ET_DIR,
    DevLicense,
    clear_cookies,
    download,
    flash_image,
    get_content,
    get_cookies,
    login,
)


log = logging.getLogger(__name__)


class FlashThread(QThread):
    flash_failed = Signal(Exception, str)
    flash_done = Signal()
    flash_progress = Signal(int)

    def __init__(
        self, bin_file: str, flash_device: CommDevice, device_name: Optional[str]
    ) -> None:
        super().__init__()
        self.bin_file = bin_file
        self.flash_device = flash_device
        self.device_name = device_name

    def run(self) -> None:
        def progress_callback(progress: int, end: bool = False) -> None:
            if end:
                self.flash_progress.emit(100)
            else:
                self.flash_progress.emit(progress)

        try:
            flash_image(
                self.bin_file,
                self.flash_device,
                device_name=self.device_name,
                progress_callback=progress_callback,
            )
            self.flash_done.emit()
        except Exception as e:
            log.error(str(e))
            self.flash_failed.emit(HandledException(e), traceback.format_exc())


class AuthThread(QThread):
    license_loaded = Signal(DevLicense)
    auth_failed = Signal()
    auth_done = Signal(tuple)  # (RequestsCookieJar, Tuple[bool, Session, Response]))

    def __init__(
        self,
        usr: str = "",
        pwd: str = "",
        do_login: bool = False,
        dev_license: Optional[DevLicense] = None,
    ) -> None:
        super().__init__()
        self.usr = usr
        self.pwd = pwd
        self.do_login = do_login
        self.dev_license = dev_license

    def run(self) -> None:
        login_succeeded = False

        if self.do_login:
            cookies = login(self.usr, self.pwd)
        else:
            cookies = get_cookies()

        try:
            if cookies is not None:
                content = get_content(cookies)
                login_succeeded = content[0]
            else:
                login_succeeded = False
        except Exception as e:
            log.error(str(e))

        if self.dev_license is not None:
            self.dev_license.load()
            self.license_loaded.emit(self.dev_license)

        if login_succeeded:
            self.auth_done.emit((cookies, content))
        else:
            if not self.do_login:
                clear_cookies()
            self.auth_failed.emit()


class BinDownloadThread(QThread):
    download_done = Signal(str, str)
    download_failed = Signal(str)

    def __init__(
        self,
        device: Optional[str],
        cookies: RequestsCookieJar,
        session: Session,
        response: Response,
    ) -> None:
        super().__init__()
        self.device = device
        self.cookies = cookies
        self.session = session
        self.response = response

    def run(self) -> None:
        if (
            self.device is None
            or self.cookies is None
            or self.session is None
            or self.response is None
        ):
            self.download_failed.emit("Failed to start download")
        else:
            with TemporaryDirectory() as temp_dir:
                try:
                    temp_file, version = download(
                        session=self.session,
                        cookies=self.cookies,
                        path=temp_dir,
                        device=self.device,
                    )

                    if temp_file is not None:
                        bin_file = str(ET_DIR / Path(temp_file).name)
                        shutil.move(temp_file, bin_file)
                        self.download_done.emit(bin_file, version)
                    else:
                        self.download_failed.emit(
                            f"Failed to download firmware for device {self.device}"
                        )
                except Exception as e:
                    self.download_failed.emit(str(e))
                finally:
                    self.session.close()



# Copyright (c) Acconeer AB, 2024
# All rights reserved

"""
The core of this module is the Epoch class, which has a type-safe[1]
builder-API to build a "migration tree" from the leaves and up.

An instance of Epoch can be seen as a node and the node data is a
python type (called 'head').
Each node keeps edges from its children to itself (called 'inbounds').
Children are also Epochs.

Here are some side-by-sides with timeline definitions and the migration tree:

=================================================================================
| Definition                                         | Migration tree           |
=================================================================================
|                                                    |                          |
|                                                    |                          |
| tm.start(int)                                      |  (h=int) <=top           |
|                                                    |                          |
|                                                    |                          |
|----------------------------------------------------|--------------------------|
|                                                    |  (h=int) <=top           |
| tm.start(int)                                      |     ^                    |
|   .load(str, lambda s: int(s), fail=...)           |     | via the lambda     |
|                                                    |     |                    |
|                                                    |  (h=str)                 |
|----------------------------------------------------|--------------------------|
|                                                    |  (h=int) <=top           |
| tm.start(int)                                      |    ^ ^                   |
|   .load(str, lambda s: int(s), fail=...)           |    | |                   |
|   .load(float, lambda s: int(s), fail=...)         |    | +-------+           |
|                                                    |    |         |           |
|                                                    |  (h=str)  (h=float)      |
|----------------------------------------------------|--------------------------|
|  tm.start(int)                                     |  (h=bytes) <=top         |
|    .load(str, lambda s: int(s), fail=...)          |     ^                    |
|    .load(float, lambda s: int(s), fail=...)        |     |                    |
|    .epoch(byte, lambda i: bytes(i), fail=...)      |     |                    |
|                                                    |  (h=int)                 |
|                                                    |    ^ ^                   |
|                                                    |    | |                   |
|                                                    |    | +-------+           |
|                                                    |    |         |           |
|                                                    |  (h=str)  (h=float)      |
|----------------------------------------------------|--------------------------|
|                                                    |  (h=bytes) <=top         |
|  tm.start(int)                                     |     ^ ^                  |
|    .load(str, lambda s: int(s), fail=...)          |     | |                  |
|    .load(float, lambda s: int(s), fail=...         |     | +------+           |
|    .epoch(byte, lambda i: bytes(i), fail=...)      |     |        |           |
|    .load(str,                                      |  (h=int)  (h=str)        |
|          lambda s: bytes(s, encoding="ascii"),     |    ^ ^                   |
|          fail=...)                                 |    | |                   |
|                                                    |    | +-------+           |
|                                                    |    |         |           |
|                                                    |  (h=str)  (h=float)      |
|----------------------------------------------------|--------------------------|

When calling 'migrate', a depth-first-search is done in the
tree to find a matching node for the type of the passed 'obj'.

When a matching Epoch is found, the "edge" function to the parent node then applied
for each level of the tree until the root is found. If the "edge" function raises
one of the Exceptions specified in 'fail', the search continues in the tree.
If an "edge" function raises and error that is not specified in 'fail', that
exception will be propagated.

If this search fails, a MigrateError is thrown, otherwise a migrated instance of
correct type is returned.

[1] type-safe practically means your editor will yell at you for using the wrong types
"""

from __future__ import annotations

import functools
import typing as t

import attrs
import attrs.validators as av
import result as r
import typing_extensions as te


_T = t.TypeVar("_T")
_R = t.TypeVar("_R")
_S = t.TypeVar("_S")
_P = te.ParamSpec("_P")

_TBE = t.TypeVar("_TBE", bound=BaseException)
_ReqCtxT = t.TypeVar("_ReqCtxT")
_NewCtxT = t.TypeVar("_NewCtxT")
_MigT = t.TypeVar("_MigT")
_HeadT = t.TypeVar("_HeadT")

Completer: te.TypeAlias = t.Callable[[t.Type[_T]], _T]
_Transform: te.TypeAlias = t.Callable[[_T, Completer[_ReqCtxT]], _R]


def _add_ignored_completer(f: t.Callable[[_T], _R]) -> _Transform[_T, te.Never, _R]:
    """Turns a regular function into a _Transform for homogenous internal representation.
    Used in the non-contextual 'load' and 'epoch'
    """
    if not callable(f):
        msg = "f must be a callable"
        raise TypeError(msg)

    def wrapper(__x: _T, _: Completer[t.Any]) -> _R:
        return f(__x)

    return wrapper


def _flip(f: t.Callable[[_S, _T], _R]) -> t.Callable[[_T, _S], _R]:
    """Flips the parameters of a 2-parameter function"""

    def wrapper(__x: _T, __y: _S) -> _R:
        return f(__y, __x)

    return wrapper


def _as_result(
    *exceptions: type[_TBE],
) -> t.Callable[
    [t.Callable[_P, _R]],
    t.Callable[_P, r.Result[_R, _TBE]],
]:
    """Extends result.as_result to accept 0 caught exceptions"""
    if exceptions:
        return r.as_result(*exceptions)

    def _decorator_factory(f: t.Callable[_P, _R]) -> t.Callable[_P, r.Result[_R, _TBE]]:
        def _decorator(*args: _P.args, **kwargs: _P.kwargs) -> r.Ok[_R]:
            return r.Ok(f(*args, **kwargs))

        return _decorator

    return _decorator_factory


def _null_completer(typ: type) -> t.Any:
    msg = f"No completion for {typ}"
    raise RuntimeError(msg)


class MigrationError(Exception):
    """This error is raised then a migration fails"""


def _inbounds_validator(_ignored1: t.Any, _ignored2: t.Any, value: t.Any) -> None:
    (transform, ancestor) = value

    if not callable(transform):
        msg = "The first element of the tuple needs to be a callable"
        raise TypeError(msg)

    if not isinstance(ancestor, Epoch):
        msg = "The second element of the tuple needs to be an Epoch"
        raise TypeError(msg)


@attrs.frozen
class Epoch(t.Generic[_HeadT, _ReqCtxT, _MigT]):
    head: type[_HeadT] = attrs.field(validator=av.instance_of(type))
    inbounds: t.Sequence[
        tuple[
            _Transform[t.Any, _ReqCtxT, r.Result[_HeadT, Exception]],
            Epoch[t.Any, t.Any, t.Any],
        ],
    ] = attrs.field(validator=av.deep_iterable(_inbounds_validator))

    def _is_supported(self, typ: type) -> bool:
        return issubclass(typ, self.head) or any(
            ancestor._is_supported(typ) for _, ancestor in self.inbounds
        )

    def load(
        self,
        src: type[_T],
        f: t.Callable[[_T], _HeadT],
        fail: t.Sequence[type[Exception]],
    ) -> Epoch[_HeadT, _ReqCtxT, _MigT | _T]:
        """Add a loading function to the current epoch"""
        new_gen = start(src)
        return Epoch(
            self.head,
            [(_as_result(*fail)(_add_ignored_completer(f)), new_gen), *self.inbounds],  # type: ignore[arg-type]
        )

    def contextual_load(
        self,
        src: type[_T],
        f: _Transform[_T, _NewCtxT, _HeadT],
        fail: t.Sequence[type[Exception]],
    ) -> Epoch[_HeadT, _ReqCtxT | _NewCtxT, _MigT | _T]:
        """
        Add a contextual loading function to the current epoch.
        `f`'s second argument should be a `Completer`
        """
        new_gen = start(src)
        return Epoch(
            self.head,
            [(_as_result(*fail)(f), new_gen), *self.inbounds],  # type: ignore[arg-type, list-item]
        )

    def epoch(
        self,
        typ: type[_T],
        f: t.Callable[[_HeadT], _T],
        fail: t.Sequence[type[Exception]],
    ) -> Epoch[_T, _ReqCtxT, _MigT | _HeadT]:
        """Append an epoch to the timeline, adding a level to the migration tree"""
        return Epoch(
            typ,
            [(_as_result(*fail)(_add_ignored_completer(f)), self)],  # type: ignore[arg-type]
        )

    def contextual_epoch(
        self,
        typ: type[_T],
        f: _Transform[_HeadT, _NewCtxT, _T],
        fail: t.Sequence[type[Exception]],
    ) -> Epoch[_T, _ReqCtxT | _NewCtxT, _MigT | _HeadT]:
        """Append a contextual epoch to the timeline, adding a level to the migration tree"""
        return Epoch(typ, [(_as_result(*fail)(f), self)])  # type: ignore[arg-type]

    def _migrate_results(
        self,
        obj: _HeadT | _MigT,
        completer: Completer[_ReqCtxT],
    ) -> t.Iterator[r.Result[_HeadT, Exception]]:
        if isinstance(obj, self.head):
            yield r.Ok(obj)
            return

        yield from (
            migrated.and_then(functools.partial(_flip(transform), completer))
            for transform, ancestor in self.inbounds
            for migrated in ancestor._migrate_results(obj, completer)
        )

    @te.overload
    def migrate(self: Epoch[_HeadT, te.Never, _MigT], obj: _HeadT | _MigT) -> _HeadT: ...

    @te.overload
    def migrate(
        self: Epoch[_HeadT, _ReqCtxT, _MigT],
        obj: _HeadT | _MigT,
        completer: Completer[_ReqCtxT],
    ) -> _HeadT: ...

    def migrate(
        self, obj: _HeadT | _MigT, completer: Completer[t.Any] = _null_completer
    ) -> _HeadT:
        """
        Migrate `obj` be the type of the latest epoch.

        raises `MigrationError` if migration was not possible
        """
        if not self._is_supported(type(obj)):
            msg = f"Cannot migrate objects of type {type(obj)}"
            raise TypeError(msg)

        for res in self._migrate_results(obj, completer):
            if isinstance(res, r.Ok):
                return res.ok_value

        msg = f"Failed during migration of object {obj!r}"
        raise MigrationError(msg)

    def nop(self) -> te.Self:
        """Does nothing."""
        return self


def start(typ: type[_T]) -> Epoch[_T, te.Never, te.Never]:
    return Epoch(head=typ, inbounds=[])
